
module Plugins.Interaction where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad.Trans
import Control.Monad.State
import Control.Monad.Reader

import qualified Data.Map as Map

import System.Directory
import System.IO    hiding (print, putStr, putStrLn)
import System.Console.GetOpt

import Syntax.Position
import Syntax.Internal
import Syntax.Values
import Syntax.Pretty
import Syntax.ConcreteToInternal

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Errors
import TypeChecker.Values.DeBruijn
import TypeChecker.Context
import TypeChecker.Monad.Context
import TypeChecker.UnfoldConstants
import TypeChecker

import Interface.Interactive
import IO.Readline

import Utils.Error
import Utils.Monads
import Utils.Unicode
import Utils.Pairs
import Utils.Names

import Plugins

plugin :: SomePlugin
plugin = mkPlugin $ 
    Plugin  { plName		= "interact"
            , plHelp            = pluginHelp
	    , plInit		= initialize
	    , plFinal		= finalize
	    , plCheck		= check
	    , plInfer		= infer
	    , plOptions		= options
	    , plDefaultOptions	= defaultOptions
	    }

type PluginState = () -- [InteractionPoint]

data InteractionPoint = IP  { ipEnv	    :: Environment
			    , ipState	    :: TCState
			    , ipGoalType    :: Type
			    , ipRange	    :: Range
			    }

pluginHelp = "Interaction plugin. Gives you an interactive prompt."

initialize :: Options -> TC PluginState
initialize _ = return ()

finalize :: Options -> PluginState -> TC ()
finalize _ _ = return ()

prompt :: InteractionPoint -> TC Value
prompt ip =
    do	liftIO $ print $ ipRange ip
	put $ ipState ip
	local (const $ ipEnv ip) $ return vTop
	it <- fresh "it"
	xs    <- map fst <$> getContext
	scope <- getScopeEnv
	withScopeEnv (addVarNames xs scope)
	    $ interaction "interact-plugin> " (cmds it) (evalTerm it)
    where
	(|>|) = (,)
	x .>. y = (x, const y)

	cmds :: Name -> [Command Value]
	cmds it =
	    [ "quit"	.>. give False ["assert-plugin()"]
	    , "give"	|>| give False
	    , "replace" |>| give True

	    , "type"	|>| \args ->
		do  typeTerm it (unwords args)
		    return Continue
	    , "try"	|>| \args ->
		do  give False args
		    return Continue

	    , "where"	.>.
		do  liftIO (print $ ipRange ip)
		    return Continue

	    , "goal"	.>.
	        do  v' <- unBruijn =<<
			    typeMapM unfoldAllConstants (ipGoalType ip)
		    liftIO $ print v'
		    return Continue

	    , "context" .>.
	        do  showContext =<< getContext
		    return Continue

	    , "rules"	.>.
		do  loc  <- concat . Map.elems . envLocalRewriteRules <$> ask
		    glob <- concat . Map.elems . tcGlobalRewriteRules <$> get
		    (loc,glob) <- unBruijn (loc,glob)
		    liftIO $
			do  unless (null glob) $
				do  putStrLn $ "Global rules:"
				    mapM_ print glob
			    unless (null loc) $
				do  putStrLn $ "Local rules:"
				    mapM_ print loc
		    return Continue

	    , "help"	.>. help it
	    , "?"	.>. help it
	    ]

	help it =
	    do	liftIO $ putStrLn $ "Commands:"
		mapM_ (\cmd -> liftIO $ putStrLn $ "  :" ++ fst cmd) (cmds it)
		return Continue

	give sharp args =
	    do	scope <- getScopeEnv
		(e,_) <- parseTermAt (rStart $ ipRange ip) scope s
		-- You had better be sure when you do replace.
		-- It will replace even if it isn't type correct!
		-- This is to make successive refinements work.
		when sharp $ liftIO $ patchFile (ipRange ip) s
		v <- checkType e $ ipGoalType ip
		return $ Return v
	    where
		s = unwords args

	showContext gamma = inContext [] $ showCtx (n - 1) $ reverse gamma
	    where
		n = length [ () | (_,Abstract _) <- gamma ]
		showCtx n ((x,Abstract t):gamma) =
		    do	t' <- unBruijn =<< typeMapM unfoldAllConstants t
			liftIO $ print $
			    text ("(@" ++ show n ++ ")")
			    <+> sep [ pretty x
				    , nest 2 $ text ":" <+> pretty t'
				    ]
			extendContext x (Abstract t) $ showCtx (n - 1) gamma
		showCtx n ((x,Defined v t):gamma) =
		    do	v' <- unfoldAllConstants v
			t' <- typeMapM unfoldConstants t
			(v'',t'') <- unBruijn (v',t')
			liftIO $ print $
			    text "(--)"
			    <+> sep [ pretty x
				    , nest 2 $ text ":" <+> pretty t''
				    , nest 2 $ text "=" <+> pretty v''
				    ]
			extendContext x (Defined v t) $ showCtx n gamma
		showCtx _ [] = return ()

patchFile :: Range -> String -> IO ()
patchFile r p =
    do	exist <- doesFileExist file
	when exist $
	    do	cont <- readFileUTF8 file
		()   <- last cont `seq` return ()   -- force reading
		writeFileUTF8 file (patch r p cont)
    where
	file = srcFile $ rStart r

	patch r p s =
	    case splitAtPos (rStart r) s of
		(hd, s') -> case splitAtPos' (rStart r) (rEnd r) s' of
		    (_, tl) -> hd ++ p ++ tl

	splitAtPos pos s = splitAtPos' (startPos file) pos s

	splitAtPos' p0 p1 [] = ([], [])
	splitAtPos' p0 p1 (x:xs)
	    | before	= (x:) -*- id $ splitAtPos' (movePos p0 x) p1 xs
	    | otherwise	= ([], x:xs)
	    where
		before = l0 < l1 || l0 == l1 && c0 < c1
		l0 = posLine p0
		l1 = posLine p1
		c0 = posCol p0
		c1 = posCol p1

check :: Options -> PluginState -> Range -> [Term] -> Type -> TC (Value, PluginState)
check o s r es t =
    do	env <- ask
	tcs <- get
	v   <- if optAssert o then checkType (CallPlugin r "assert" []) t
			      else prompt (IP env tcs t r)
	return (v,())

infer :: Options -> PluginState -> Range -> [Term] -> TC (Value, Type, PluginState)
infer _ _ _ _ = failure $ customError "You can't infer the type of an interaction point"

data Options = Opts { optAssert :: Bool }

options :: PluginOptions Options
options = simpleOptions
	    [ Option [] ["not"] (NoArg $ \o -> return $ o { optAssert = True })
			"don't interact (act as the assert-plugin)"
	    ]

defaultOptions = Opts False

