-- Normalizing If-Expressions
-- Author:: Andreas Abel
-- Created:: 2004-12-07

-- import Bool
-- Defines the booleans and some functions over them.
-- Author :: Ulf Norell
-- Created:: 2004-12-06

-- The booleans -----------------------------------------------------------
data Bool : Set where
  false	: Bool
  true	: Bool

-- Equality ---------------------------------------------------------------
infix 4 `EqBool`
EqBool = Eq Bool

-- Elimination principle --------------------------------------------------
boolCase : (P : Bool -> Prop) ->
	    P false -> P true	->
	    (b : Bool) -> P b
boolCase P pF pT false	= pF
boolCase P pF pT true	= pT

-- Turning Bools into Props -----------------------------------------------
T : Bool -> Prop
T true = True
T false = False

TTrue : (b : Bool) -> Prop
TTrue b = T b ==> b `EqBool` true

Ttrue : (b : Bool) -> TTrue b
Ttrue b = fol-plugin(boolCase TTrue)

TFalse : (b : Bool) -> Prop
TFalse b = Not (T b) ==> b `EqBool` false

Tfalse : (b : Bool) -> TFalse b
Tfalse b = fol-plugin(boolCase TFalse)

-- If-then-else -----------------------------------------------------------
ifD : (C : Bool -> Set) -> (b:Bool) -> C true -> C false -> C b
ifD C true x y	= x
ifD C false x y	= y

-- If we define if in terms of ifD
-- we get problems with lambdas in FOL terms
if : (A:Set) -> Bool -> A -> A -> A
if A true x y = x
if A false x y = y

IfBranch : (A : Set) -> (x : Bool) -> Prop
IfBranch A x = (a, b : A) -> Eq A (if A x a b) a \/ Eq A (if A x a b) b

ifBranch : (A : Set) -> (x : Bool) -> IfBranch A x
ifBranch A = fol-plugin (boolCase (IfBranch A))

-- Logical operations -----------------------------------------------------
infixr 6 &&
infixr 5 ||

not : Bool -> Bool
not true  = false
not false = true

(&&) : Bool -> Bool -> Bool
x && y = if Bool x y false

(||) : Bool -> Bool -> Bool
x || y = if Bool x true y

-- Properties
AndT : (x : Bool) -> Prop
AndT x = (y : Bool) -> T (x && y) <=> T x /\ T y

andT : (x : Bool) -> AndT x
andT x y = fol-plugin (boolCase AndT)

AndAssoc : (x : Bool) -> Prop
AndAssoc x = (y, z : Bool) -> (x && y) && z `EqBool` x && (y && z)

andAssoc : (x : Bool) -> AndAssoc x
andAssoc = fol-plugin (boolCase AndAssoc)

TrueOrFalse : Bool -> Prop
TrueOrFalse x = T x \/ T (not x)
trueOrFalse : (x : Bool) -> TrueOrFalse x
trueOrFalse = fol-plugin (boolCase TrueOrFalse)

DeMorgan1 : (x:Bool) -> Prop
DeMorgan1 x = (y:Bool) -> not (x && y) `EqBool` not x || not y

deMorgan1 : (x:Bool) -> DeMorgan1 x
deMorgan1 = fol-plugin(boolCase DeMorgan1)


-- If expressions over variables in some set A

data IfExp (A:Set) : Set where
    Bit : Bool -> IfExp A
    Var : A -> IfExp A
    If  : IfExp A -> IfExp A -> IfExp A -> IfExp A

-- Semantics

eval : (A:Set) -> (A -> Bool) -> (IfExp A -> Bool)
eval A f (Bit b) = b
eval A f (Var a) = f a
eval A f (If c t e) = if Bool (eval A f c) (eval A f t) (eval A f e)

-- Normal if expressions

data NIfExp (A:Set) : Set where
    NBit : Bool -> NIfExp A
    NVar : A -> NIfExp A
    NIf  : A -> NIfExp A -> NIfExp A -> NIfExp A

-- Embedding into IfExp

emb : (A:Set) -> NIfExp A -> IfExp A
emb A (NBit b) = Bit b
emb A (NVar a) = Var a
emb A (NIf a t e) = If (Var a) (emb A t) (emb A e)

-- Normalization of if expressions

-- composition of normal expressions

nif : (A:Set) -> NIfExp A -> NIfExp A -> NIfExp A -> NIfExp A
nif A (NBit true) t e = t
nif A (NBit false) t e = e
nif A (NVar a) t e = NIf a t e
nif A (NIf a b c) t e = NIf a (nif A b t e) (nif A c t e)

-- normalization

norm : (A:Set) -> IfExp A -> NIfExp A
norm A (Bit b) = NBit b
norm A (Var a) = NVar a
norm A (If c t e) = nif A (norm A c) (norm A t) (norm A e)

-- Soundness of normalization ----------------------------------------

-- Induction for IfExp

indIfExp : (A:Set) -> (P : IfExp A -> Prop) ->
            ((b:Bool) -> P (Bit b)) ->
            ((a:A) -> P (Var a)) ->
            ((c,t,e : IfExp A) -> P c -> P t -> P e -> P (If c t e)) ->
            (e : IfExp A) -> P e
indIfExp A P bit var if (Bit b) = bit b
indIfExp A P bit var if (Var a) = var a
indIfExp A P bit var if (If c t e)  = if c t e (indIfExp A P bit var if c)
                                               (indIfExp A P bit var if t)  
                                               (indIfExp A P bit var if e)

-- Induction for NIfExp

indNIfExp : (A:Set) -> (P : NIfExp A -> Prop) ->
             ((b:Bool) -> P (NBit b)) ->
             ((a:A) -> P (NVar a)) ->
             ((a:A) -> (t,e : NIfExp A) -> P t -> P e -> P (NIf a t e)) ->
             (e : NIfExp A) -> P e
indNIfExp A P bit var if (NBit b) = bit b
indNIfExp A P bit var if (NVar a) = var a
indNIfExp A P bit var if (NIf a t e) = if a t e
                                          (indNIfExp A P bit var if t)  
                                          (indNIfExp A P bit var if e)

-- Lemma:: rotation for if

Rot_if : (A:Set) -> (b:Bool) -> Prop
Rot_if A b = (ct,ce:Bool) -> (t,e:A) -> 
           Eq A (if A (if Bool b ct ce) t e)
                (if A b (if A ct t e) (if A ce t e))

rot_if : (A:Set) -> (b:Bool) -> Rot_if A b
rot_if A = boolCase (Rot_if A)
           (fol-plugin (Bool, if)) 
           (fol-plugin (Bool, if)) 

-- Lemma:: soundness of nif

{- Lemma:: eval f (nif c t e) == eval f (If c t e)
          (emb implicit)
          
Proof:: By induction c.

Case NIf:: 
  LHS     eval f (nif (If a ct ce) t e)
       == eval f (If a (nif ct t e) (nif ce t e))
       == if (f a) (eval f (nif ct t e)) (eval f (nif ce t e))
  RHS     eval f (If (If a ct ce) t e)
       == if (eval f (If a ct ce)) (eval f t) (eval f e)
       == if (if (f a) (eval f ct) (eval f ce)) (eval f t) (eval f t)
-}

{-
Sound_nif : (A:Set) -> (f:A -> Bool) -> (t,e : NIfExp A) ->
             NIfExp A -> Prop
Sound_nif = \ A f t e c -> 
              Eq Bool (eval A f (If (emb A c) (emb A t) (emb A e))) 
                      (eval A f (emb A (nif A c t e)))

sound_nif  : (A:Set) -> (f:A -> Bool) -> (t,e : NIfExp A) ->
              (c : NIfExp A) -> Sound_nif A f t e c
sound_nif A f t e =  indNIfExp A (Sound_nif A f t e)
                      (boolCase (\ b -> Sound_nif A f t e (NBit b))
                        fol-plugin(Bool,NIfExp,eval,emb,nif)
                        fol-plugin(Bool,NIfExp,eval,emb,nif))
                      fol-plugin(Bool,NIfExp,eval,emb,nif)
                      fol-plugin(Bool,NIfExp,eval,emb,nif)
-}

Sound_nif : (A:Set) -> (f:A -> Bool) -> NIfExp A -> Prop
Sound_nif = \ A f c -> (t,e : NIfExp A) ->
              Eq Bool (eval A f (If (emb A c) (emb A t) (emb A e))) 
                      (eval A f (emb A (nif A c t e)))

sound_nif  : (A:Set) -> (f:A -> Bool) -> (c : NIfExp A) ->
              Sound_nif A f c
sound_nif A f = indNIfExp A (Sound_nif A f)
                  (boolCase (\ b -> Sound_nif A f (NBit b))
                    fol-plugin(Bool,NIfExp,eval,emb,nif)
                    fol-plugin(Bool,NIfExp,eval,emb,nif))
                  fol-plugin(Bool,NIfExp,eval,emb,nif)
                  (\ a t e ih_t ih_e ->
                     fol-plugin(Bool,NIfExp,eval,emb,nif,ih_t,ih_e,rot_if))

{-
sound_nif  : (A:Set) -> (f:A -> Bool) -> (c,t,e : NIfExp A) ->
              Eq Bool (eval A f (If (emb A c) (emb A t) (emb A e))) 
                      (eval A f (emb A (nif A c t e)))
sound_nif A f c t e = infNIfExp A fol-plugin(Bool,NIfExp,eval,emb,nif)
-}

-- Theorem:: soundness of norm

sound_norm : (A:Set) -> (f:A -> Bool) -> (e:IfExp A) ->
              Eq Bool (eval A f e) 
                      (eval A f (emb A (norm A e)))